'''
Definition of the simple 2D lidar problem, used to create the KF model.

The system state is its location and velocity (x,y,vx,vy).
The dynamics model is a constant-velocity motion.
The observation model is a single sensor with known location that observes the location (x,y).
The goal is to predict either the state (x,y,vx,vy) or the location (x,y).

Written by Ido Greenberg, 2021
'''

import numpy as np
import torch

def get_F():
    # x,y,vx,vy
    return torch.tensor([
        [1,0,1,0],
        [0,1,0,1],
        [0,0,1,0],
        [0,0,0,1]
    ], dtype=torch.double)

def get_H():
    # x,y,vx,vy -> x,y
    return torch.tensor([
        [1,0,0,0],
        [0,1,0,0],
    ], dtype=torch.double)

def initial_observation_to_state(z):
    # x,y -> x,y,vx,vy
    return torch.cat((z, torch.zeros(2,dtype=torch.double)))

def loss_fun(location_only=True):
    # MSE over either the location or the whole state (location+velocity)
    if location_only:
        return lambda pred, x: ((pred[:2]-x[:2])**2).sum()
    return lambda pred, x: ((pred-x)**2).sum()

def model_args():
    return dict(
        dim_x = 4,
        dim_z = 2,
        init_z2x = initial_observation_to_state,
        F = get_F(),
        H = get_H(),
        loss_fun = loss_fun(),
    )



'''
Optimizing and testing the Kalman Filter from data of trajectories with both observations and system states.

Contents:
- TRAIN a KF from data: both noise-estimation and gradient-based optimization are supported.
- TEST a trained model.
- ANALYSIS of the test results.

Written by Ido Greenberg, 2021
'''

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

from warnings import warn
from time import time

import torch
from torch import optim

from . import utils

###################   TRAIN   ###################

def train_models(models, X, Y, **kwargs):
    '''Run train() iteratively for every model.'''
    res_per_iter, res_per_sample = pd.DataFrame(), pd.DataFrame()
    for m in models:
        r1, r2 = train(m, X, Y, **kwargs)
        res_per_iter = pd.concat((res_per_iter, r1))
        res_per_sample = pd.concat((res_per_sample, r2))
    return res_per_iter, res_per_sample

def train(model, X, Y, split_data=None, p_valid=0.15, n_epochs=1, batch_size=10,
          lr=1e-2, lr_decay=0.5, lr_decay_freq=150, optimizer=optim.Adam, weight_decay=0.0,
          loss_after_pred=False, log_interval=300, reset_model=True, noise_estimation_initialization=True,
          best_valid_loss=np.inf, verbose=1, valid_hor=8, to_save=True, save_best=True, **kwargs):
    '''
    Training of the Kalman Filter from data of trajectories that includes both observations (X) and system states (Y).

    If the input model is configured not to be optimized, then this function only triggers the noise-estimation tuning
    from the given data.

    :param model: An instance of OKF to train.
    :param X: Observations data - list of numpy arrays of shape (n_time_steps, observaton dimension).
    :param Y: System states data - list of numpy arrays of shape (n_time_steps, state dimension).
    :param split_data: Train and validation data (Xt,Yt,Xv,Yv). If None, these are generated from X,Y. Default = None.
    :param p_valid: Percent of data used for validation (if split_data is None). Default = 0.15.
    :param n_epochs: Epochs to train (how many times to sample each trajectory). Default = 1.
    :param batch_size: Number of trajectories per training iteration. Default = 10.
    :param lr: Initial learning rate. Default = 1e-2.
    :param lr_decay:      Every lr_decay_freq iterations, lr is multiplied by lr_decay. Default = 0.5.
    :param lr_decay_freq: Every lr_decay_freq iterations, lr is multiplied by lr_decay. If None, frequency is once per
                          epoch. Default = 150.
    :param optimizer: Optimizer constructor to use. Default = optim.Adam.
    :param weight_decay: Weight decay of the parameters in the optimizer. Default = 0. Decay>0 may harm stability.
    :param loss_after_pred: Whether to calculate the loss after the prediction step or after the update (filter) step.
                            Default = False (after update step).
    :param log_interval: Validation loss is calculated every log_interval training samples. Default = 300.
    :param reset_model: Whether to reset the model before the training begins. Default = True.
    :param noise_estimation_initialization: Whether to initialize the weights using noise estimation. Default = True.
    :param best_valid_loss: Best validation loss until now. Used for continuing training. Default = inf.
    :param verbose: Verbosity level. Default = 2.
    :param valid_hor: After valid_hor validation calculations in a row w/o improvement, the training stops. Default = 8.
    :param to_save: Whether to save the resulting model. If string, saves the model under this name. If True, saves
                    under model.model_name. Default = True.
    :param save_best: Whether to save the last model or the one with the best validation loss. Default = True.
    :param kwargs: Arguments to pass to train_step().
    :return: Two data frames describing the results: losses per iteration; and final validation loss per data sample.
    '''

    # pre-processing
    Xt0, Yt0, Xv, Yv = split_train_valid(X, Y, p_valid) if split_data is None else split_data
    n_samples = len(Xt0)
    n_batches = n_samples // batch_size
    log_interval = log_interval // batch_size
    if lr_decay_freq is None: lr_decay_freq = n_batches

    if reset_model:
        model.reset_model()

    if noise_estimation_initialization or not model.optimize:
        # estimate noise covariance
        model.estimate_noise(Y,X)

    if model.optimize:
        # initialize variables
        early_stop = False
        no_improvement_seq = 0
        e = 0
        b = 0
        # train monitor
        t = []
        losses = []
        RMSE = []
        # validation monitor
        t_valid = []
        losses_valid = []
        RMSE_valid = []

        # initialize model and optimizer
        model.train()
        if weight_decay != 0:
            warn('Note: weight decay is known to cause instabilities in the tracker training.')
        params = model.parameters()
        o = optimizer(params, lr=lr, weight_decay=weight_decay)
        sched = optim.lr_scheduler.StepLR(o, step_size=lr_decay_freq, gamma=lr_decay)

        if verbose >= 1:
            print(f'\nTraining {model.model_name:s}:')
            print(f'samples={len(Xt0):d}(t)+{len(Xv):d}(v)={len(X):d}; batch_size={batch_size:d}; ' + \
                  f'iterations={n_epochs}(e)x{n_batches}(b)={n_epochs * n_batches:d}.')

        # train
        T0 = time()
        for e in range(n_epochs):
            t0 = time()

            # shuffle batches
            ids = np.arange(len(Xt0))
            np.random.shuffle(ids)
            Xt = [Xt0[i] for i in ids]
            Yt = [Yt0[i] for i in ids]

            for b in range(n_batches):
                tt = e * n_batches + b
                x = [Xt[b*batch_size+i] for i in range(batch_size)]
                y = [Yt[b*batch_size+i] for i in range(batch_size)]

                loss_batch = train_step(x, y, model, o, loss_after_pred=loss_after_pred, **kwargs)

                t.append(tt+1)
                losses.append(loss_batch)
                RMSE.append(np.sqrt(loss_batch))

                if log_interval > 0 and ((tt % log_interval == 0) or tt==n_epochs*n_batches-1):
                    # calculate validation loss
                    loss_batch = test_model(model, Xv, Yv, detailed=False, loss_after_pred=loss_after_pred)
                    model.train()

                    t_valid.append(tt+1)
                    losses_valid.append(loss_batch)
                    RMSE_valid.append(np.sqrt(loss_batch))

                    if verbose >= 2:
                        print(f'\t[{model.model_name:s}] {e + 1:02d}.{b + 1:04d}/{n_epochs:02d}.{n_batches:04d}:\t' + \
                              f'train_RMSE={RMSE[-1]:.2f}, valid_RMSE={RMSE_valid[-1]:.2f}   |   {time() - t0:.0f} [s]')

                    if len(losses_valid)>1 and np.min(losses_valid[:-1])<best_valid_loss:
                        best_valid_loss = np.min(losses_valid[:-1])
                    improved = loss_batch < best_valid_loss
                    if improved:
                        no_improvement_seq = 0
                        if to_save and save_best:
                            model.save_model(to_save if isinstance(to_save, str) else None)
                    else:
                        no_improvement_seq += 1
                    if no_improvement_seq >= valid_hor:
                        early_stop = True
                        break

                # update lr
                sched.step()

            if verbose >= 2:
                print(f'[{model.model_name:s}] Epoch {e + 1}/{n_epochs} ({time() - t0:.0f} [s])')

            if early_stop:
                break

        if verbose >= 1:
            print_train_summary(n_epochs, n_batches, early_stop, e, b, best_valid_loss, T0, model.model_name)

        # summarize losses per iteration
        res = pd.concat((
            pd.DataFrame(dict(model=len(t)*[model.model_name], t=t, group=len(t)*['train'],
                              loss=losses, RMSE=RMSE)),
            pd.DataFrame(dict(model=len(t_valid)*[model.model_name], t=t_valid, group=len(t_valid)*['valid'],
                              loss=losses_valid, RMSE=RMSE_valid))
        )).copy()

    else:
        res = pd.DataFrame({})

    # save model
    if to_save:
        if save_best and model.optimize:
            # in this case saving was already done during the training - every time validation loss achieved new record.
            # need to reload the last saved model, which is the best one over all validations.
            model.load_model(to_save if isinstance(to_save, str) else None)
        else:
            model.save_model(to_save if isinstance(to_save, str) else None)

    # detailed test over the validation data
    res_valid = test_model(model, Xv, Yv, detailed=True, loss_after_pred=loss_after_pred).copy()

    return res, res_valid

def train_step(X, Y, model, optimizer, clip=1, loss_after_pred=False, optimize_per_target=False):
    '''
    A single training step - one batch of trajectories.

    For most parameters - see train() documentation.
    :param clip: Clip the gradients to this value. Default = 1.
    :param optimize_per_target: When calculating the loss, assign the same weight to every trajectory (target) rather
                                than every sample (time-step). That is, don't give more weight to longer trajectories.
    :return: The loss of this training step.
    '''

    # assign weights to errors (uniform over time-steps or uniform over targets)
    targets_lengths = np.array([len(x) for x in X])
    targets_weights = 1/targets_lengths if optimize_per_target else np.ones(len(X))
    targets_weights = targets_weights / np.sum(targets_weights*targets_lengths)

    optimizer.zero_grad()
    tot_loss = torch.tensor(0.)
    for x,y,w in zip(X,Y,targets_weights):
        model.init_state()
        for t in range(len(x)):
            xx = x[t,:]
            yy = y[t,:]

            loss = None
            model.predict()
            if loss_after_pred:
                loss = model.loss_fun(model.x, torch.tensor(yy)) if t>0 else torch.tensor(0.)

            model.update(xx)
            if not loss_after_pred:
                loss = model.loss_fun(model.x, torch.tensor(yy))

            tot_loss = tot_loss + w * loss

    tot_loss.backward()
    if clip:
        torch.nn.utils.clip_grad_norm_(model.parameters(), clip)
    optimizer.step()

    return tot_loss.item()

def split_train_valid(X, Y, p=0.15, seed=9):
    n_valid = int(np.round(p*len(X)))
    np.random.seed(seed)
    ids_valid = set(list(np.random.choice(np.arange(len(X)), n_valid, replace=False)))
    Xt = [x for i,x in enumerate(X) if i not in ids_valid]
    Yt = [x for i,x in enumerate(Y) if i not in ids_valid]
    Xv = [x for i,x in enumerate(X) if i in ids_valid]
    Yv = [x for i,x in enumerate(Y) if i in ids_valid]
    return Xt, Yt, Xv, Yv

def print_train_summary(n_epochs, n_batches, early_stop, epoch, i_batch, valid_loss, T0, tit):
    print(f'[{tit:s}] Training done ({time() - T0:.0f} [s])')
    if early_stop:
        print(
            f'\tbest valid loss: {valid_loss:.0f};\tearly stopping:\t{epoch + 1:d}.{i_batch + 1:03d}/{n_epochs:d}.{n_batches:03d} ({100 * (epoch*n_batches + i_batch + 1) / (n_epochs * n_batches):.0f}%)')
    else:
        print(
            f'\tbest valid loss: {valid_loss:.0f};\tno early stopping:\t{n_epochs:d} epochs, {n_batches:d} batches, {n_epochs * n_batches:d} total iterations.')

###################   TEST   ###################

def test_model(model, X, Y, detailed=False, loss_fun=None, loss_after_pred=False, count_base=0, verbose=0):
    # detailed = Whether to return a detailed data-frame or just the final loss.
    with torch.no_grad():
        model.eval()
        if loss_fun is None: loss_fun = model.loss_fun
        # per-step data
        targets = []
        times = []
        losses = []
        SE = []
        AE = []
        # per-batch data
        tot_loss = 0

        count = 0
        t0 = time()
        if verbose >= 1:
            print(f'\nTesting {model.model_name:s}:')
        for tar, (XX, YY) in enumerate(zip(X, Y)):
            model.init_state()
            for t in range(len(XX)):
                count += 1
                x = XX[t,:]
                y = YY[t,:]

                model.predict()
                if loss_after_pred:
                    loss = loss_fun(model.x, y) if t>0 else torch.tensor(0.)

                model.update(x)
                if not loss_after_pred:
                    loss = loss_fun(model.x, y)

                loss = loss.item()
                tot_loss += loss

                if detailed:
                    targets.append(count_base+tar)
                    times.append(t)
                    SE.append(loss)
                    AE.append(np.sqrt(loss))
                    losses.append(loss)

        if verbose >= 1:
            print(f'done.\t({time()-t0:.0f} [s])')
        if detailed:
            return pd.DataFrame(dict(
                model = len(times) * [model.model_name],
                target = targets,
                t = times,
                SE = SE,
                AE = AE,
                loss=losses,
            ))
        tot_loss /= count
        return tot_loss

###################   ANALYSIS   ###################

def analyze_test_results(res):
    models = np.unique(res.model)
    axs = utils.Axes(6, 4, axsize=(5, 3.8))
    a = 0

    for m in models:
        mean = res[res.model==m].AE.mean()
        utils.plot_quantiles(res[res.model==m].AE.values, showmeans=True, ax=axs[a], label=f'{m}  (mean={mean:.1f})',
                             linewidth=2, means_args=dict(linewidth=2))
    axs[a].legend(fontsize=15)
    axs.labs(a, 'quantile [%]', 'error', fontsize=15)
    a += 1

    sns.boxplot(data=res, x='model', y='AE', showmeans=True, showfliers=False, ax=axs[a],
                meanprops=dict(marker='o', markerfacecolor='w', markeredgecolor='r', markersize=10))
    axs.labs(a, 'model', 'error', fontsize=16)
    axs[a].set_xticklabels(axs[a].get_xticklabels(), fontsize=16)
    a += 1

    sns.barplot(data=res, x='model', y='SE', capsize=.07, ci=99, ax=axs[a])
    axs.labs(a, 'model', 'error^2 (99% confidence)', fontsize=16)
    a += 1

    sns.lineplot(data=res, x='t', hue='model', y='AE', ax=axs[a])
    axs.labs(a, 'time-step', 'error', fontsize=16)
    a += 1

    axs[a].axhline(0, color='k')
    for i1,m1 in enumerate(models):
        for m2 in models[i1+1:]:
            delta = res[res.model == m2].AE.values - res[res.model==m1].AE.values
            zval = np.mean(delta) / np.std(delta) * np.sqrt(len(delta))
            utils.plot_quantiles(delta, showmeans=True, ax=axs[a], label=f'{m2}-{m1} (z-val={zval:.1f})',
                                 linewidth=2, means_args=dict(linewidth=2))
    axs[a].legend(fontsize=15)
    axs.labs(a, 'quantile [%]', 'error(model_2) - error(model_1)',
             title=f'Models comparison in pairs\n(sample = time-step)', fontsize=13)
    a += 1

    axs[a].axhline(0, color='k')
    for i1,m1 in enumerate(models):
        for m2 in models[i1+1:]:
            delta = (res[res.model == m2].groupby('target').apply(lambda d: np.sqrt(d.SE.mean())) -
                     res[res.model == m1].groupby('target').apply(lambda d: np.sqrt(d.SE.mean())) )
            zval = np.mean(delta) / np.std(delta) * np.sqrt(len(delta))
            utils.plot_quantiles(delta, showmeans=True, ax=axs[a], label=f'{m2}-{m1} (z-val={zval:.1f})',
                                 linewidth=2, means_args=dict(linewidth=2))
    axs[a].legend(fontsize=15)
    axs.labs(a, 'target quantile [%]', 'error(model_2) - error(model_1)',
             title=f'Models comparison in pairs\n(sample = target)', fontsize=13)
    a += 1

    plt.tight_layout()
    return axs

def display_tracking(models, X, Y, n=4, t_min=0, xdim=0, ydim=1, plot_after_pred=False, show_observations=False):
    axs = utils.Axes(n, 4, axsize=(5, 4))
    colors = ['r', 'b', 'g', 'y']
    preds = {}
    for i in range(n):
        ax = axs[i]
        preds[i] = []
        XX = X[i]
        # plot target
        ax.plot(Y[i][t_min:, xdim], Y[i][t_min:, ydim], 'k.-', label='ground-truth')
        # plot observations
        if show_observations:
            ax.plot(X[i][t_min:, xdim], X[i][t_min:, ydim], 'm.', label='observation')
        ax.plot(Y[i][t_min:t_min+1, xdim], Y[i][t_min:t_min+1, ydim], 'k>', markersize=10)
        # plot models outputs
        for j, m in enumerate(models):
            preds[i].append([])
            m.init_state()
            for t in range(len(XX)):
                x = XX[t, :]
                m.predict()
                if plot_after_pred and t>0:
                    preds[i][j].append(m.x.detach().numpy())
                m.update(x)
                if not plot_after_pred:
                    preds[i][j].append(m.x.detach().numpy())
            preds[i][j] = np.stack(preds[i][j])  # prediction[target][model]
            ax.plot(preds[i][j][t_min:, xdim], preds[i][j][t_min:, ydim], colors[j%len(colors)], linewidth=1.5,
                    label=f'{m.model_name}')

        axs.labs(i, None, None, f'{t_min:d}<t<{len(XX):d}', fontsize=14)
        ax.legend(fontsize=14)

    plt.tight_layout()
    return axs



'''
General utils (figures, etc.).

Written by Ido Greenberg, 2020
'''

import gc
from time import sleep
from warnings import warn
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import torch

DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

def get_quantiles(X, Q=100, assume_sorted=False):
    if not assume_sorted:
        X = sorted(X)
    N = len(X)
    try:
        n = len(Q)
        Q = np.array(Q)
    except:
        Q = np.arange(0,1+1e-6,1/Q)
        Q[-1] = 1
        n = len(Q)
    x = [X[int(q*(N-1))] for q in Q]
    return x, Q

def plot_quantiles(x, ax=None, Q=100, showmeans=False, plot_mean=False, means_args=None, **kwargs):
    if plot_mean: showmeans = True # backward compatibility
    if ax is None: ax = Axes(1,1)[0]
    m = np.mean(x)
    x, q = get_quantiles(x, Q)
    h = ax.plot(100*q, x, '-', **kwargs)
    if showmeans:
        if means_args is None: means_args = {}
        ax.axhline(m, linestyle='--', color=h[0].get_color(), **means_args)
    return ax

def labels(ax, xlab=None, ylab=None, title=None, fontsize=12):
    if isinstance(fontsize, int):
        fontsize = 3*[fontsize]
    if xlab is not None:
        ax.set_xlabel(xlab, fontsize=fontsize[0])
    if ylab is not None:
        ax.set_ylabel(ylab, fontsize=fontsize[1])
    if title is not None:
        ax.set_title(title, fontsize=fontsize[2])

def fontsize(ax, labs=16, ticks=12, leg=None, draw=True, wait=0):
    if wait:
        sleep(wait)
    if draw:
        plt.draw()
    if labs is not None:
        if not isinstance(labs, (tuple,list)):
            labs = 3*[labs]
            ax.set_xlabel(ax.get_xlabel(), fontsize=labs[0])
            ax.set_ylabel(ax.get_ylabel(), fontsize=labs[1])
            ax.set_title(ax.get_title(), fontsize=labs[2])
    if ticks is not None:
        if not isinstance(ticks, (tuple,list)):
            ticks = 2*[ticks]
            ax.set_xticklabels(ax.get_xticklabels(), fontsize=ticks[0])
            ax.set_yticklabels(ax.get_yticklabels(), fontsize=ticks[1])
    if leg is not None:
        ax.legend(fontsize=leg)

class Axes:
    def __init__(self, N, W=2, axsize=(5,3.5), grid=1, fontsize=13):
        self.fontsize = fontsize
        self.N = N
        self.W = W
        self.H = int(np.ceil(N/W))
        self.axs = plt.subplots(self.H, self.W, figsize=(self.W*axsize[0], self.H*axsize[1]))[1]
        for i in range(self.N):
            if grid == 1:
                self[i].grid(color='k', linestyle=':', linewidth=0.3)
            elif grid ==2:
                self[i].grid()
        for i in range(self.N, self.W*self.H):
            self[i].axis('off')

    def __len__(self):
        return self.N

    def __getitem__(self, item):
        if self.H == 1 and self.W == 1:
            return self.axs
        elif self.H == 1 or self.W == 1:
            return self.axs[item]
        return self.axs[item//self.W, item % self.W]

    def labs(self, item, *args, **kwargs):
        if 'fontsize' not in kwargs:
            kwargs['fontsize'] = self.fontsize
        labels(self[item], *args, **kwargs)

def to_device(model, tensor_list):
    if model is not None:
        model.to(DEVICE)
    for i,tns in enumerate(tensor_list):
        if tns is not None:
            if type(tns) is not torch.Tensor:
                tns = torch.from_numpy(tns)
            tensor_list[i] = tns.to(DEVICE)
    return tensor_list

def clean_device(tensor_list):
    for i in reversed(list(range(len(tensor_list)))):
        del tensor_list[i]
    gc.collect()
    torch.cuda.empty_cache()

def update_dict(base_dict, dict_to_add=None, force=False, copy=False):
    if copy:
        base_dict = base_dict.copy()
    if dict_to_add is not None:
        for k, v in dict_to_add.items():
            if force or k not in base_dict:
                base_dict[k] = v
    return base_dict

def pd_merge_cols(dd, cols_to_merge, values_cols=('Y',), case_col='case', cases_names=None, cols_to_keep=None):
    if isinstance(cols_to_merge[0], str):   cols_to_merge = [cols_to_merge]
    if isinstance(values_cols, str):        values_cols = [values_cols]
    if cases_names is None:                 cases_names = cols_to_merge[0]
    if cols_to_keep is None:
        all_cols_to_merge = [c for cols in cols_to_merge for c in cols]
        cols_to_keep = [c for c in dd.columns if c not in all_cols_to_merge]

    n_merges = len(cols_to_merge)
    n_to_merge = len(cols_to_merge[0])
    d = pd.DataFrame()
    for c in cols_to_keep:
        d[c] = n_to_merge * list(dd[c].values)

    d[case_col] = np.repeat(cases_names, len(dd))
    for cols, values_col in zip(cols_to_merge, values_cols):
        d[values_col] = np.concatenate([dd[c].values for c in cols])

    return d

def overwrite_legend(ax, fun, filter=None, fontsize=12):
    h, cur_labs = ax.get_legend_handles_labels()
    if filter is not None:
        tmp = [(hh, l) for hh, l in zip(h, cur_labs) if filter(l)]
        h = [tt[0] for tt in tmp]
        cur_labs = [tt[1] for tt in tmp]
    labs = [fun(l) for l in cur_labs]
    ax.legend(h, labs, fontsize=fontsize)

def compare_with_ref(dd, y, x, ref=None, x_ord=None, hue=None, sort_hue=True, ax=None, font=14, **kwargs):
    if ax is None: ax = Axes(1,1,axsize=(6,4))[0]
    if x_ord is None: x_ord = sorted(np.unique(dd[x]))
    if ref is None: ref = x_ord[0]
    x_ord = [xx for xx in x_ord if xx!=ref]

    dd = dd.copy()
    dd.loc[:,'delta'] = 0.
    ids0 = dd[x] == ref
    for xx in x_ord:
        ids = dd[x] == xx
        if np.sum(ids) != np.sum(ids0):
            # cant run pairwise comparison (unpaired comparison is not implemented)
            return
        dd.loc[ids,'delta'] = dd.loc[ids,y].values - dd.loc[ids0,y].values

    dd = dd[dd[x]!=ref]
    if len(dd) == 0:
        warn('A single x - nothing to compare to.')
        return

    sns.barplot(data=dd, x=x, y='delta', hue=hue, hue_order=sorted(dd[hue]) if hue and sort_hue else None,
                ax=ax, capsize=.1, **kwargs)
    if hue is not None:
        ax.legend(fontsize=12)
    ylab = f'{y:s} difference'
    labels(ax, 'Alternative', ylab, f'Alternative vs. {ref:s}', fontsize=font)
    plt.draw()
    fontsize(ax, font, 12, None if hue is None else 12)
    if len(x_ord) > 4:
        ax.set_xticklabels(ax.get_xticklabels(), rotation=25)

def qgroups(x, nbins=5, apply_labs=True):
    qs, _ = get_quantiles(x, nbins)
    qs = np.array(qs)
    qs[-1] += 1
    g = [int(np.sum(xx>=qs))-1 for xx in x]
    qs[-1] -= 1
    v = [f'{qs[gg]:.0f}-{qs[gg+1]:.0f}' for gg in g] if apply_labs else g
    sorted_v = [vv for _,vv in sorted(zip(g,v))]
    ids = np.unique(sorted_v, return_index=True)[1]
    v_labs = [sorted_v[i] for i in sorted(ids)]
    return g, v, v_labs

def compare_quantiles(dd, x, y, hue=None, fac=None, xbins=5, hbins=5, fbins=4, ci=95, box=False,
                      mean=np.mean, copy=True, axs=None, axs_args=None, **kwargs):
    if copy: dd = dd.copy()
    # prepare groups
    xx = x+'_tmp'
    relabel = False
    if xbins and len(np.unique(dd[x]))>2.5*xbins:
        g, v, relabel = qgroups(dd[x], xbins)
        dd[xx] = g
    else:
        dd[xx] = dd[x]
    hues = None
    if hue:
        hh = hue + '_tmp'
        dd[hh] = dd[hue]
        hues = np.unique(dd[hh])
        if len(hues) > 1.7*hbins:
            _, dd[hue], hues = qgroups(dd[hh], hbins)
    facs = ff = None
    if fac:
        ff = fac + '_tmp'
        facs = np.unique(dd[fac])
        if len(facs) > 2*fbins:
            _, dd[ff], facs = qgroups(dd[fac], fbins)
        else:
            dd[ff] = dd[fac]
        nf = len(facs)
    else:
        nf = 1

    # prepare axes
    if axs is None:
        if axs_args is None: axs_args = {}
        axs_def_args = dict(W=3, axsize=(7,5), fontsize=15)
        axs_args = update_dict(axs_args, axs_def_args)
        axs = Axes(nf, **axs_args)

    # plot
    pd.options.mode.chained_assignment = None  # default='warn'
    for a in range(nf):
        ax = axs[a]
        d = dd[dd[ff]==facs[a]] if fac else dd
        curr_hues = [h for h in hues] if hue else None
        if mean is not None and hue:
            for i,h in enumerate(curr_hues):
                m = mean(d[d[hue]==h][y])
                hh = f'{h:s} ({m:.0f})'
                d.loc[d[hue]==h, hue] = hh
                curr_hues[i] = hh
        if box:
            sns.boxplot(data=d, x=xx, y=y, hue=hue, hue_order=curr_hues, ax=ax, showmeans=True, **kwargs)
        else:
            sns.lineplot(data=d, x=xx, y=y, hue=hue, hue_order=curr_hues, ax=ax, ci=ci, **kwargs)
        if relabel:
            ax.set_xticks(np.arange(len(relabel)))
            ax.set_xticklabels(relabel, fontsize=12, rotation=20)
        labels(axs[a], x, y, f'{fac}: {facs[a]}' if fac else None, fontsize=15)
        axs[a].legend(title=hue, fontsize=12)
    pd.options.mode.chained_assignment = 'warn'
    plt.tight_layout()

    return axs



'''Optimized Kalman Filter: optimize your KF parameters from data of system-states and observations.'''
from okf.model import OKF
from okf.optimizer import train_models, train, split_train_valid, test_model, analyze_test_results, display_tracking
from okf import utils
from okf import example


'''
This module implements the Optimized Kalman Filter (OKF) class.
The OKF is similar to any KF implementation, but the parameters are torch variables such that they can be optimized.

The OKF class includes:
- Interfaces:
    - A constructor (see documentation in __init__()).
    - init_state(): Initialize the model before a new sequence of observations.
    - reset_model(): Reset the model parameters (Q,R).
    - save_model(), load_model().
- Methods to apply the model (to be run iteratively):
    - predict(): Advance a single time-step and update the state.
    - update(): Process a new observation and update the state.
- Tuning the model parameters:
    - estimate_noise(): Set the parameters to be the sample covariance matrices of the noise in the given data.
    - Optimization of the parameters wrt a loss function is implemented in a separate module.
- Methods to observe the model: get_Q(), get_R(), display_params().

Written by Ido Greenberg, 2021
'''

import types
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

import torch
import torch.nn as nn
from torch import matmul as mp

from . import utils

class OKF(nn.Module):
    def __init__(self, dim_x, dim_z, F, H, model_name='OKF', P0=1e3, Q0=1, R0=1,
                 init_z2x=None, loss_fun=None, optimize=True, model_files_path='models/'):
        '''
        A model of KF whose parameters (Q,R) are pytorch tensors and can be optimized wrt a loss function.

        Under the KF assumptions, such optimization (wrt the MSE of the model-predictions) is equivalent to simply
        calculate the sample covariance matrices of the noise (given corresponding data). However, the KF assumptions
        do not usually hold in practical problems - which is sometimes goes unnoticed. Optimization can obtain better
        accuracy in such cases.

        :param dim_x: System-state ("hidden-state") dimension [int].
        :param dim_z: Observation (measurement) dimension [int].
        :param F: Process (dynamics) model [pytorch tensor of type double and shape (dim_x,dim_x) OR fun(x, z)
                  that returns such a tensor].
        :param H: Observation (measurement) model [pytorch tensor of type double and shape (dim_z,dim_x) OR fun(x, z)
                  that returns such a tensor].
        :param model_name: Model name [str].
        :param P0: The initial value of the uncertainty matrix P, used to initialize P every new trajectory. If scalar,
                   the initial P is P0*eye(dim_x) [numeric OR pytorch tensor with type double and shape (dim_x,dim_x);
                   default=1e3].
        :param Q0: The initial value of the process-noise covariance matrix Q, from which the optimization begins.
                   If scalar, it is used as a scale drawing the initial matrix randomly [positive numeric OR pytorch
                   tensor with type double and shape (dim_x,dim_x); default=1; only used if optimize==True].
        :param R0: The initial value of the observation-noise covariance matrix R, from which the optimization begins.
                   If scalar, it is used as a scale drawing the initial matrix randomly [positive numeric OR pytorch
                   tensor with type double and shape (dim_z,dim_z); default=1; only used if optimize==True].
        :param init_z2x: A function that receives the first observation and returns the first estimate of the state
                         [fun(z); default=identity function; if dim_x!=dim_z, another function must be specified].
        :param loss_fun: Loss function to optimize [fun(predicted_x, true_x); default=MSE; only used if optimize==True].
        :param optimize: Whether to tune the parameters Q,R by optimization or using the standard sample covariance
                         matrices of the noise.
        :param model_files_path: Directory path to save the model in [str].
        '''
        nn.Module.__init__(self)
        self.model_name = model_name
        self.base_path = model_files_path
        self.optimize = optimize
        self.dim_x = dim_x
        self.dim_z = dim_z

        self.F = F
        self.H = H
        self.is_H_fun = isinstance(self.H, types.FunctionType)
        self.is_F_fun = isinstance(self.F, types.FunctionType)

        self.P0 = P0
        self.Q0 = Q0
        self.R0 = R0
        self.Q_D, self.Q_L, self.R_D, self.R_L = 4 * [None]
        self.reset_model()

        self.z2x = init_z2x
        if self.z2x is None:
            if self.dim_x == self.dim_z:
                raise ValueError('Whenever dim_x!=dim_z, the init_z2x mapping must be specified explicitly.')
            self.z2x = lambda x: x

        self.loss_fun = loss_fun
        if self.loss_fun is None:
            self.loss_fun = lambda pred,x: ((pred-x)**2).sum()

        self.x = None
        self.z = None
        self.P = None
        self.init_state()

    def init_state(self):
        '''Initialize the estimate (x,P) and the observation (z) before a new sequence of observations (trajectory).'''
        self.x = self.dim_x * [None]
        self.z = self.dim_z * [None]
        self.P = self.P0 * torch.eye(self.dim_x, dtype=torch.double)

    def reset_model(self):
        '''Reset the model parameters (Q,R).'''
        # Q
        if isinstance(self.Q0, torch.Tensor) and len(self.Q0.shape):
            # given as a 2D tensor
            Q_D, Q_L = OKF.encode_SPD(self.Q0)
        else:
            # given as a scale for randomization
            Q_D = (self.Q0 * (0.5 + torch.rand(self.dim_x, dtype=torch.double))).log()
            Q_L = self.Q0/5 * torch.randn(self.dim_x * (self.dim_x-1) // 2, dtype=torch.double)
        # R
        if isinstance(self.R0, torch.Tensor) and len(self.R0.shape):
            # given as a 2D tensor
            R_D, R_L = OKF.encode_SPD(self.R0)
        else:
            # given as a scale for randomization
            R_D = (self.R0 * (0.5 + torch.rand(self.dim_z, dtype=torch.double))).log()
            R_L = self.R0/5 * torch.randn(self.dim_z * (self.dim_z-1) // 2, dtype=torch.double)

        if self.optimize:
            self.Q_D = nn.Parameter(Q_D, requires_grad=True)
            self.Q_L = nn.Parameter(Q_L, requires_grad=True)
            self.R_D = nn.Parameter(R_D, requires_grad=True)
            self.R_L = nn.Parameter(R_L, requires_grad=True)
        else:
            self.Q_D, self.Q_L, self.R_D, self.R_L = Q_D, Q_L, R_D, R_L

    def save_model(self, fname=None, base_path=None, assert_suffices=True):
        fpath = self.get_model_path(fname, base_path, assert_suffices)
        if self.optimize:
            # module state dict
            torch.save(self.state_dict(), fpath)
        else:
            # save tensors
            torch.save((self.Q_D, self.Q_L, self.R_D, self.R_L), fpath)

    def load_model(self, fname=None, base_path=None, assert_suffices=True):
        fpath = self.get_model_path(fname, base_path, assert_suffices)
        if self.optimize:
            # module state dict
            self.load_state_dict(torch.load(fpath))
        else:
            # saved tensors
            self.Q_D, self.Q_L, self.R_D, self.R_L = torch.load(fpath)

    def get_model_path(self, fname=None, base_path=None, assert_suffices=True):
        if base_path is None: base_path = self.base_path
        if fname is None: fname = self.model_name
        if assert_suffices:
            if base_path[-1] != '/': base_path += '/'
            if self.optimize:
                if not fname.endswith('.m'): fname += '.m'
            else:
                if not fname.endswith('.noise'): fname += '.noise'
        return base_path + fname

    def predict(self):
        if self.x[0] is None:
            # state has not yet been initialized (probably no observation has yet been processed)
            return
        F = self.F(self.x, self.z) if self.is_F_fun else self.F
        Q = OKF.get_SPD(self.Q_D, self.Q_L)
        self.x = mp(F, self.x)
        self.P = mp(mp(F, self.P), F.T) + Q

    def update(self, z):
        # get observation
        self.z = torch.tensor(z)
        H = self.H(self.x, self.z) if self.is_H_fun else self.H

        # get update operators
        R = OKF.get_SPD(self.R_D, self.R_L)
        Ht = H.T
        PHt = mp(self.P, Ht)
        self.S = mp(H, PHt) + R
        K = mp(PHt, self.S.inverse())

        # update P
        I_KH = torch.eye(self.P.shape[0]) - mp(K, H)
        self.P = mp(mp(I_KH, self.P), I_KH.T) + mp(mp(K, R), K.T)  # equivalent to the standart formula
                                                                   # but more numerically-stable
        self.P = 0.5 * (self.P + self.P.T)  # force P to be symmetric

        # update x
        if self.x[0] is not None:
            self.x = self.x + mp(K, self.z - mp(H, self.x))
        else:
            self.x = self.z2x(self.z)

    @staticmethod
    def get_SPD(D, L):
        '''Convert log-diagonal entries [n] and below [n*(n-1)/2] into a SPD matrix [n^2].'''
        n = len(D)
        A = D.exp().diag() # fill diagonal
        ids = torch.tril_indices(n, n, -1)
        A[ids[0, :], ids[1, :]] = L # fill below-diagonal
        return mp(A, A.T)

    @staticmethod
    def encode_SPD(A, eps=1e-6):
        '''Apply Cholesky decomposition to A and return the log-diagonal entires [n] and below [n*(n-1)/2].'''
        n = A.shape[0]
        A = torch.cholesky(A+eps*torch.eye(n))
        D = A.diag()
        D = D.log()
        ids = torch.tril_indices(n,n,-1)
        L = A[ids[0,:],ids[1,:]]
        return D, L

    def estimate_noise(self, X, Z):
        '''
        Tune the KF by noise estimation:
        Set the parameters Q,R to be the sample covariance matrices of the noise in the given data.

        :param X: a list of targets states. X[i] = numpy array of type double and shape (n_time_steps(i), dim_x).
        :param Z: a list of targets observations. Z[i] = numpy array of type double and shape (n_time_steps(i), dim_z).
        '''

        # Q = Cov[F*x_t - x_{t+1}]
        X1 = torch.cat([torch.tensor(x[:-1]) for x in X], dim=0) # x_t
        X2 = torch.cat([torch.tensor(x[1:])  for x in X], dim=0) # x_{t+1}
        if self.is_F_fun:
            F = [self.F(torch.tensor(x), torch.tensor(z)) for x, z in zip(X, Z)]
            Fx1 = np.concatenate([mp(f, torch.tensor(x).T).T.detach().numpy() for x, f in zip(X, F)], axis=0)
        else:
            Fx1 = mp(self.F, X1.T).T  # F*x_t
        Q = torch.tensor(np.cov((Fx1-X2).T.detach().numpy()))

        # R = Cov[z_t - H*x_t]
        Z = np.concatenate(Z, axis=0)
        H = [self.H(torch.tensor(x), torch.tensor(z)) for x,z in zip(X,Z)] \
            if self.is_H_fun else len(X)*[self.H]
        Hx = np.concatenate([mp(h,torch.tensor(x).T).T.detach().numpy() for x,h in zip(X,H)], axis=0)
        delta = Z - Hx
        R = torch.tensor(np.cov((delta).T))

        # Cholesky parameterization
        Q_D, Q_L = OKF.encode_SPD(Q)
        R_D, R_L = OKF.encode_SPD(R)
        if self.optimize:
            with torch.no_grad():
                self.Q_D.copy_(Q_D)
                self.Q_L.copy_(Q_L)
                self.R_D.copy_(R_D)
                self.R_L.copy_(R_L)
        else:
            self.Q_D, self.Q_L = Q_D, Q_L
            self.R_D, self.R_L = R_D, R_L

    def get_Q(self, to_numpy=True):
        A = OKF.get_SPD(self.Q_D, self.Q_L)
        if to_numpy: A = A.detach().numpy()
        return A

    def get_R(self, to_numpy=True):
        A = OKF.get_SPD(self.R_D, self.R_L)
        if to_numpy: A = A.detach().numpy()
        return A

    def display_params(self, n_digits=0, fontsize=15, axsize=(4.5, 3.5)):
        axs = utils.Axes(2, 2, axsize=axsize)
        for i,A in enumerate([self.get_Q(), self.get_R()]):
            h = sns.heatmap(A, annot=True, fmt=f'.{n_digits:d}f', cmap="Reds", ax=axs[i],
                            annot_kws=None if fontsize is None else dict(size=fontsize))
            h.xaxis.set_ticks_position("top")
        axs.labs(0, title=f'[{self.model_name}] Q', fontsize=fontsize+2)
        axs.labs(1, title=f'[{self.model_name}] R', fontsize=fontsize+2)
        plt.tight_layout()
        return axs




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Optimized Kalman Filter",
    version="0.0.1",
    license='MIT',
    author="Ido Greenberg",
    description="Optimization of a Kalman Filter from data of states and their observations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    url="https://github.com/ido90/Optimized-Kalman-Filter",
    keywords = ["OKF", "Optimized Kalman Filter", "Kalman Filter"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.5',
    py_modules=["okf"],
    install_requires=["torch"]
)



'''
A simulator of data for the simple 2D lidar problem, used as an example to use for KF optimization.

The system state is its location and velocity (x,y,vx,vy).
The targets trajectories consist of alternating straight intervals and turns, with random accelerations.
The observations are noisy samples of the target location (x,y), as measured by a single sensor with known location.

Written by Ido Greenberg, 2021
'''

import pickle as pkl
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from .. import utils


############   DATA API   ############

def load_data(fpath='data/simple_lidar_data.pkl'):
    with open(fpath, 'rb') as fd:
        X, Z = pkl.load(fd)
    return X, Z

def get_trainable_data(X, Z):
    '''Convert the list of data-frames into numpy arrays.'''
    return [x.values.astype(np.double) for x in X], [z.values.astype(np.double) for z in Z]

def display_data(X, Z):
    axs = utils.Axes(8, 4, (5, 3.5))

    utils.plot_quantiles([len(x) for x in X], axs[0], showmeans=True)
    axs.labs(0, 'Quantile [%]', 'Number of time steps', f'Lengths distribution over {len(X):d} trajectories')

    for i in range(6):
        axs[1].plot(X[i]['x'], X[i]['y'], '-', label=str(i))
    axs[1].set_title('A sample of trajectories', fontsize=14)
    axs[1].legend()

    for i in range(6):
        ax = axs[2+i]
        ax.plot(X[i]['x'], X[i]['y'], 'k.-', label='state')
        ax.plot(Z[i]['zx'], Z[i]['zy'], 'r.', label='observation')
        axs.labs(2+i, 'x', 'y', f'Target {i}')
        ax.legend()

    plt.tight_layout()


############   SIMULATOR   ############

def rand_range(a,b):
    return (b - a) * np.random.rand() + a

def rand_range_sym(d):
    return d * (2*np.random.rand()-1)

def simulate_data(n_targets=1000, x0=600, y0=600, v0=(30,90),
                  n_intervals=(4,7), int_len=(5,9), ar_sigma=2, at_sigma=15,
                  noise_r=2, noise_t=1.5*np.pi/180, fpath='data/simple_lidar_data.pkl'):
    X, Z = [], []
    for i in range(n_targets):
        x, y, vx, vy, zx, zy = [],[],[],[],[],[]
        # init state
        x.append(rand_range_sym(x0))
        y.append(rand_range_sym(y0))
        vtheta = rand_range(0,2*np.pi)
        v = rand_range(*v0)
        vx.append(v*np.cos(vtheta))
        vy.append(v*np.sin(vtheta))
        r = np.linalg.norm((x[-1],y[-1])) + np.random.normal(0, noise_r)
        theta = np.arctan2(y[-1], x[-1]) + np.random.normal(0, noise_t)
        zx.append(r*np.cos(theta))
        zy.append(r*np.sin(theta))
        # intervals
        n_ints = np.random.randint(*n_intervals)
        for interval in range(n_ints):
            interval_len = np.random.randint(*int_len)
            ar = np.random.normal(0, ar_sigma)
            at = rand_range_sym(at_sigma)
            for t in range(interval_len):
                # simulate motion
                v = np.linalg.norm((vx[-1],vy[-1]))
                ax = ar*vx[-1]/v - at*vy[-1]/v
                ay = ar*vy[-1]/v + at*vx[-1]/v
                x.append(x[-1]+vx[-1]+0.5*ax)
                y.append(y[-1]+vy[-1]+0.5*ay)
                vx.append(vx[-1]+ax)
                vy.append(vy[-1]+ay)
                # simulate observation
                r = np.linalg.norm((x[-1],y[-1])) + np.random.normal(0, noise_r)
                theta = np.arctan2(y[-1], x[-1]) + np.random.normal(0, noise_t)
                zx.append(r*np.cos(theta))
                zy.append(r*np.sin(theta))
        # save target
        XX = pd.DataFrame(dict(
            x = x,
            y = y,
            vx = vx,
            vy = vy,
        ))
        ZZ = pd.DataFrame(dict(
            zx = zx,
            zy = zy,
        ))
        X.append(XX)
        Z.append(ZZ)
    if fpath:
        with open(fpath, 'wb') as fd:
            pkl.dump((X,Z), fd)
    return X, Z
